var signer = require("ApiGateway/signer.js");
var pkg = require('../../../../package.json');
var iamAPI = require("./iam_api.json");
var url = require("url");


/**
 * 从pkg文件中读出iam的endpoint
 * @param pkg
 * @returns {*}
 */
function getIAMEndpoint(pkg) {
    try {
        return pkg['access']['endpoint']
    } catch (err) {
        logger.warn("get the iam info failed from package.json.");
        return "";
    }
}

const DOMAIN_CACHE_FILE = global.appDataPath + "domainInfo";

/**
 * 负责与iam连接，管理底层通信
 */
class IAMConnector {
    constructor() {
        this.endpoint = getIAMEndpoint(pkg);  //package.json中可配置一个默认url
        this.ak = undefined;
        this.sk = undefined;
        this.agent = undefined;
        this.domainCache = {};    //缓存domain_id
        this.userCache = []; //[{domain_id:"123","name":"jianghcangle","id":"1"},{domain_id:"123","name":"1111","id":"2"}];    //useCache信息,是个数组，每一项是一个{domain_id,name,id}
        this._loadDomainCache();
    }

    _loadDomainCache() {
        var fs = require("fs");
        var self = this;
        fs.readFile(DOMAIN_CACHE_FILE, "utf8", function (err, data) {
                if (err) {
                    logger.error("read DOMAIN_CACHE_FILE failed.", DOMAIN_CACHE_FILE, err);
                    return;
                }
                try {
                    var cache = JSON.parse(data);
                    self.domainCache = cache;
                } catch (error) {
                    logger.error("parse DOMAIN_CACHE_FILE failed. ", DOMAIN_CACHE_FILE, data, error);
                    return;
                }
            }
        );
    }

    _saveDomainCache() {
        var fs = require("fs");
        var domainStr = JSON.stringify(this.domainCache);
        fs.writeFile(DOMAIN_CACHE_FILE, domainStr, function (err) {
            if (err) {
                logger.error("save domain info failed.", DOMAIN_CACHE_FILE, domainStr);
            }
        })
    }


    /**
     * 解析endpoint，从中得到host,port信息，并返回http调用库
     * @param endpoint
     * @param httpOption
     * @returns {Object}
     * @private
     */
    _loadEndpoint(endpoint, httpOption) {
        if (!endpoint) {
            endpoint = this.endpoint;
        }
        if (endpoint.startsWith("http")) {
            let urlObj = url.parse(endpoint);
            httpOption.host = urlObj.host;
        }
        var http = endpoint.startsWith("https") ? require("https") : require("http");
        return http;
    }


    /**
     * 如果切换账户，需要重新加载一下
     * @param accountInfo
     */
    loadUserInfo(accountInfo) {
        if (accountInfo) {
            this.ak = accountInfo.accessKey;
            this.sk = accountInfo.secretKey;
        }
    }

    /**
     * 利用ak和sk进行签名
     * @param ak
     * @param sk
     * @param httpOption
     */
    _signAKSK(httpOption) {
        let sig = new signer.Signer();
        sig.AppKey = this.ak;
        sig.AppSecret = this.sk;
        let r = new signer.HttpRequest();
        r.host = httpOption.host;
        r.method = httpOption.method.toUpperCase();
        r.uri = httpOption.uri;

        if (httpOption.query) {
            r.query = httpOption.query;
        }
        if (httpOption.body) {
            r.body = httpOption.body;
        }
        r.headers = httpOption.headers;
        return sig.Sign(r);
    }

    request(operation, httpOption, callback) {
        var operation = iamAPI.operations[operation]
        operation.headers = operation.headers ? operation.headers : {};
        httpOption.headers = Object.assign(httpOption.headers, operation.headers);
        httpOption.method = operation.method;
        httpOption.uri = operation.path;
        var http = this._loadEndpoint(operation.endpoint, httpOption);

        var signObj = this._signAKSK(httpOption);
        signObj.agent = this.agent;
        signObj.rejectUnauthorized = false;
        var clientRequest = http.request(signObj, function (resp) {
            let chunks = [];
            resp.on('data', (chunk) => {
                chunks.push(chunk);
            });
            resp.on('end', () => {
                callback(null, Buffer.concat(chunks).toString("utf-8"));
            });
            resp.on("error", (err) => {
                callback(err, null);
            })
        });
        if (httpOption.body) {
            clientRequest.write(httpOption.body);
        }
        clientRequest.end();
    }

    loadAgent(agent) {
        if (agent) {
            this.agent = agent;
        } else {
            this.agent = undefined;
        }
    }

    /**
     * 查询ak sk所属租户下所有用户信息
     *
     * @param domainID ak sk所代表用户的domainid
     * @param callback   (err,data)   data:{users:[userInfo]} ,userInfo={domain_id,id,name}
     */
    queryUsers(domainID, callback) {
        var self = this;
        if (typeof(domainID) === "function") {
            callback = domainID;
            domainID = undefined;
        }
        if (!domainID) {
            domainID = this.getDomainID();
        }
        self.request("getUserList", {"headers": {"X-Domain-Id": domainID}}, function (err, data) {
            if (data) {
                data = JSON.parse(data)
                if (data.error || data.message) {
                    self.resetDomainID();
                    return callback(data.error || new Error(data.message), null);
                }
                self.userCache = data.users ? data.users : [];
                self.setDomainID(domainID);
                self._saveDomainCache();
                return callback(err, data);
            }
            if (err) {
                self.resetDomainID();
            }
            return callback(err, data);
        });
    }
    queryCDN(uri,callback) {
        var self = this;
        var option = {"domain_origin":uri}
        self.request("getCDNList",{"headers": {"x-domain-origin": uri},query:option}, function (err, data) {
            if (data) {
                data = JSON.parse(data)
                if (data.error || data.message) {
                    return callback(data.error || new Error(data.error), null);
                } else {
                    global.logger.info("查询数据", data,"===========")
                    return callback(err, data);
                }
                if (err) {
                    global.logger.info("查询报错", err)
                }
                return callback(err, data);
            }
        });
    }
    createCDND(uri,userEndpoint,Type,callback) {
        var self = this;
        var option = {
            domain: {
                domain_name: userEndpoint,
                business_type: Type,
                sources: [{
                    ip_or_domain: uri,
                    origin_type: "obs_bucket",
                    active_standby: 1
                }]
            }
        }
        option = JSON.stringify(option);
        self.request("createCDNDomain",{"headers": {"x-domain-origin": uri},body:option}, function (err, data) {
            if(err){
                global.logger.info("createCDND err",err)
            }
            global.logger.info("创建返回",data)
            if(data){
                if (data.error || data.message) {
                    return callback(data.error || new Error(data.error), null);
                } else {
                    return callback(err, data);
                }
            }
        })
    }
    getDomainID() {
        if (!this.domainCache[global.currentUserName]) {
            this.domainCache[global.currentUserName] = global.curOnwerId;
        }
        return this.domainCache[global.currentUserName] ? this.domainCache[global.currentUserName] : "";
    }
    readCDN(){
        var fs = require("fs");
        if(fs.existsSync(global.appDataPath + "CDNConfig")){
            var CDNConfig = fs.readFileSync(global.appDataPath + "CDNConfig", "utf-8");
            var b = new Base64();
            var base64Str = b.decode(CDNConfig);
            CDNConfig = JSON.parse(base64Str);
            return CDNConfig;
        }
    }
    setDomainID(domainID) {
        this.domainCache[global.currentUserName] = domainID;
    }

    //如果查询失败，则去除cache中的当前用户的domainID
    resetDomainID() {
        delete this.domainCache[global.currentUserName];
    }
}
function Base64() {
    var _keyStr = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
    this.encode = function (input) {
        var output = "";
        var chr1, chr2, chr3, enc1, enc2, enc3, enc4;
        var i = 0;
        input = _utf8_encode(input);
        while (i < input.length) {
            chr1 = input.charCodeAt(i++);
            chr2 = input.charCodeAt(i++);
            chr3 = input.charCodeAt(i++);
            enc1 = chr1 >> 2;
            enc2 = ((chr1 & 3) << 4) | (chr2 >> 4);
            enc3 = ((chr2 & 15) << 2) | (chr3 >> 6);
            enc4 = chr3 & 63;
            if (isNaN(chr2)) {
                enc3 = enc4 = 64;
            } else if (isNaN(chr3)) {
                enc4 = 64;
            }
            output = output +
                _keyStr.charAt(enc1) + _keyStr.charAt(enc2) +
                _keyStr.charAt(enc3) + _keyStr.charAt(enc4);
        }
        return output;
    }

// public method for decoding
    this.decode = function (input) {
        var output = "";
        var chr1, chr2, chr3;
        var enc1, enc2, enc3, enc4;
        var i = 0;
        input = input.replace(/[^A-Za-z0-9\+\/\=]/g, "");
        while (i < input.length) {
            enc1 = _keyStr.indexOf(input.charAt(i++));
            enc2 = _keyStr.indexOf(input.charAt(i++));
            enc3 = _keyStr.indexOf(input.charAt(i++));
            enc4 = _keyStr.indexOf(input.charAt(i++));
            chr1 = (enc1 << 2) | (enc2 >> 4);
            chr2 = ((enc2 & 15) << 4) | (enc3 >> 2);
            chr3 = ((enc3 & 3) << 6) | enc4;
            output = output + String.fromCharCode(chr1);
            if (enc3 != 64) {
                output = output + String.fromCharCode(chr2);
            }
            if (enc4 != 64) {
                output = output + String.fromCharCode(chr3);
            }
        }
        output = _utf8_decode(output);
        return output;
    }
    // private method for UTF-8 encoding
    _utf8_encode = function (inputStr) {
        inputStr = inputStr.replace(/\r\n/g, "\n");
        var utftext = "";
        for (var n = 0; n < inputStr.length; n++) {
            var c = inputStr.charCodeAt(n);
            if (c < 128) {
                utftext += String.fromCharCode(c);
            } else if ((c > 127) && (c < 2048)) {
                utftext += String.fromCharCode((c >> 6) | 192);
                utftext += String.fromCharCode((c & 63) | 128);
            } else {
                utftext += String.fromCharCode((c >> 12) | 224);
                utftext += String.fromCharCode(((c >> 6) & 63) | 128);
                utftext += String.fromCharCode((c & 63) | 128);
            }
        }
        return utftext;
    }
    // private method for UTF-8 decoding
    _utf8_decode = function (utftext) {
        var string = "";
        var i = 0;
        var c = c1 = c2 = 0;
        while (i < utftext.length) {
            c = utftext.charCodeAt(i);
            if (c < 128) {
                string += String.fromCharCode(c);
                i++;
            } else if ((c > 191) && (c < 224)) {
                c2 = utftext.charCodeAt(i + 1);
                string += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
                i += 2;
            } else {
                c2 = utftext.charCodeAt(i + 1);
                c3 = utftext.charCodeAt(i + 2);
                string += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
                i += 3;
            }
        }
        return string;
    }
}

var accessConnector = new IAMConnector();

module.exports = accessConnector;




