/**
 * new
 */
var AWS = require('../core');
var fs = require('fs');
var removeArrayItem = require("../../../../js/obsUtils").removeArrayItem;
var process = require('process');

var DataCache = function () {
    this.partNumbersArray = [];  //保存段号
    this.downloadCaches = new Map();   //key:段号,value:{buffers: 下载的buffer数组,length:buffer数组的总字节长度}
};

DataCache.prototype.push = function (partNumber, data) {
    if (!this.downloadCaches.has(partNumber)) {
        this.partNumbersArray.push(partNumber);
        var partCache = {"buffers": [data], "length": data.length}
        this.downloadCaches.set(partNumber, partCache);
    } else {
        var partCache = this.downloadCaches.get(partNumber);
        partCache["buffers"].push(data);
        partCache['length'] = partCache['length'] + data.length;
    }
};

/**
 * 从缓存中删除一个段的下载信息
 * @param partNumber
 */
DataCache.prototype.cleanPart = function (partNumber) {
    this.downloadCaches.delete(partNumber);
    var index = this.partNumbersArray.indexOf(partNumber);
    if (index >= 0) {
        removeArrayItem(this.partNumbersArray, index);
    }
}

DataCache.prototype.shift = function () {
    var partNumber = this.partNumbersArray.shift();
    if (!partNumber) {
        return undefined;
    }
    var result = this.downloadCaches.get(partNumber);
    this.downloadCaches.delete(partNumber);

    return {
        "partNumber": partNumber,
        "length": result['length'],
        "data": Buffer.concat(result['buffers'], result['length'])
    };
};

DataCache.prototype.size = function () {
    return this.downloadCaches.size;
}


AWS.S3.ManagedDownload = AWS.util.inherit({

    constructor: function ManagedDownload(options) {
        var self = this;
        AWS.SequentialExecutor.call(self);

        self.callback = null;   //跑完后的回调函数
        self.parts = {};            //保存已启动但是未完成的request。
        self.completeMap = {};  //保存已下载的段信息{key:parNumber,value:file handle position}
        self.nextChunkStart = 0;   //分段指针，表示下一段应该从哪里开始下载  // from 0
        self.totalPartNumber = 0; //总分段数
        self.fd = null;               //下载文件的本地句柄
        self.isStop = false;

        self.dataCache = new DataCache();  //从服务器读到的数据cache.
        self.isDownloadOver = false;    //整个下载是否已经结束。此值为true并不一定代表数据已经完全写入到本地硬盘
        self.isWriting = false;   //判断当前写入文件任务是否正在进行.
        self.configure(options);
    },

    configure: function configure(options) {
        this.partSize = options.partSize;
        this.queueSize = options.queueSize;

        var total = 0;
        if (options.completeMap) {
            this.completeMap = options.completeMap;
            for (var i in options.completeMap) {
                total = total + options.completeMap[i] - (i - 1) * options.partSize;
            }
        }
        this.totalDownloadedBytes = total;  //已下载的字节数

        this.service = options.service;
        this.bindServiceObject(options.params);
        this.key = options.params.Key;

        this.size = this.service.config.params.size;
        this.localpath = this.service.config.params.localpath;
    },

    send: function (callback) {
        var self = this;
        self.callback = callback || function (err) {
            if (err)
                throw err;
        };
        global.logger.info("start download..............................key and querySize and size:", self.key, self.queueSize, self.size);
        global.logger.info("[download start] totalDownloadedBytes ", self.totalDownloadedBytes);
        var p = self.prepareDownload();
        p.then(() => {
            var tasks = [];
            for (var index = 0; index < self.queueSize; index++) {
                tasks[index] = self.createChunkTask();
            }
            return Promise.all(tasks);
        }).then(() => {
            self.isDownloadOver = true;
            if (!self.isWriting) {
                self.cleanup();
            }
        }).catch((err) => {
            self.cleanup(err);
        });
    },

    /**
     * 下载前的准备工作,创建或者打开本地待下载文件
     * @returns {Promise<any>}
     */
    prepareDownload: function () {
        var self = this;
        var p = new Promise((resolve) => {
            fs.exists(self.localpath, (exist) => {
                resolve(exist);
            });
        });
        return p.then((exist) => {
            var flag = "r+";
            if (!exist) {   //如果下载文件不存在，需要reset一下原来的下载信息,重新开始
                flag = "w+";
                self.completeMap = {};
                self.totalDownloadedBytes = 0;
            }
            return new Promise((resolve, reject) => {
                fs.open(self.localpath, flag, (err, fd) => {
                    if (err) {
                        reject(err);
                    } else {
                        self.fd = fd;
                        resolve(self);
                    }
                })
            });
        });

    },

    /**
     * 创建分段下载任务并执行。
     * @returns  返回一个代表分段任务的promise，如果分段任务分配完毕，则返回一个resolve态的promise
     */
    createChunkTask: function () {
        var self = this;
        if (self.isStop) {
            return Promise.reject(new Error("the dowload has already stopped." + self.key));
        }
        if (self.nextChunkStart == self.size) {  //段已分配完毕,没有必要再起下载任务了,等待现有任务执行完毕即可
            global.logger.info("[download]the assignment for chunk to task has finish.", self.key);
            return Promise.resolve();  // 任务结束
        }
        if (self.dataCache.size() >= self.queueSize * 2) {  //磁盘写不完了，缓一下(很难出现这种情况)
            global.logger.info("[download]warn The speed of writing disk < network transport:", self.key, self.isWriting, self.dataCache.maxBufferLength);
            return new Promise((resolve) => {
                setTimeout(resolve, 1000);
            }).then(() => {
                return self.createChunkTask();
            });
        }
        var start = self.nextChunkStart;
        var end = (self.nextChunkStart + self.partSize < self.size) ? self.nextChunkStart + self.partSize - 1 : self.size - 1;
        self.nextChunkStart = end + 1;

        var partNumber = ++self.totalPartNumber;  //段号，从1开始

        if (self.completeMap[partNumber]) {
            //如果在完成列表中，则遍历列表
            while(self.completeMap[partNumber]) {
                if (self.completeMap[partNumber] > end) { // 如果该段已经完成，移动start，end，next指针
                    self.nextChunkStart = end + 1;
                    if (self.nextChunkStart === self.size){
                        return Promise.resolve();
                    }
                    start = self.nextChunkStart;
                    end = (self.nextChunkStart + self.partSize < self.size) ? self.nextChunkStart + self.partSize - 1 : self.size - 1;
                    partNumber++;
                    self.totalPartNumber++;
                    if (!self.completeMap[partNumber]) { //如果表里到列表最后，需要将新的分段插入列表中,并更新next的指针
                        self.completeMap[partNumber] = start;
                        self.nextChunkStart = end + 1;
                        break;
                    }
                } else { //如果该段还未下载完，需要更新start指针，以及next指针;
                    start = self.completeMap[partNumber];
                    self.nextChunkStart = end + 1;
                    break;
                }
            }
        } else {
            self.completeMap[partNumber] = start;
        }

        // if (self.completeMap[partNumber]) {
        //
        //     var position = self.completeMap[partNumber];
        //     logger.info("[download] continue part information,partNumber,position,end", partNumber, position, end);
        //     if (position > end) {
        //        return self.createChunkTask();
        //     } else {
        //         start = position;
        //     }
        // } else {
        //     self.completeMap[partNumber] = start;
        // }
        var partParams = {Range: "bytes=" + start + "-" + end};
        var params = {
            Bucket : global.Bucket,
            Key : self.key,
            Policy:global.Policy,
            AWSAccessKeyId:global.AccessKeyId,
            Signature:global.Signature,
            Range: "bytes=" + start + "-" + end
        };
        var p = new Promise((resolve, reject) => {
            if(global.shareDownload){
                var req = self.service.getObjectByUrl(params);
            }else{
                var req = self.service.getObject(partParams);
            }
            req._lastDownloadedBytes = 0;
            self.parts[partNumber] = req;   //保存一个活动的req句柄
            let taskDownloadSize = end - start + 1;   //本次下载任务应该下载的大小
            let currentSize=0;   //当前本任务的下载大小
            req.on('httpData', function (data, resp) {
                //如果是重定向什么的，不做任何操作
                var statusCode = resp.request && resp.request.response && resp.request.response.httpResponse && resp.request.response.httpResponse.statusCode;
                if (statusCode > 300) {
                    global.logger.warn('httpData statusCode:', resp.request.response.httpResponse.statusCode);
                    return;
                }
                currentSize = currentSize + data.length;
                if (currentSize > taskDownloadSize) {   //如果实际下载量大于了本段预计下载量，应该马上结束,避免数据污染后一段
                    let message = "download sec size > partSize.  " + currentSize + " " + start + " " + end;
                    let err = self.triggerNetWorkingError(message, partNumber);
                    return reject(err);
                }
                self.dataCache.push(partNumber, data);
                if (!self.isWriting)  //不可在多个异步任务中同时写文件
                    self.writeFile();
            });
            //本段下载结束(注意只是http传输结束，但是不一定写入了本地文件)
            req.on("success", function () {
                global.logger.info("[download] s3.request http Done.", self.key, partNumber, partParams.Range);
                delete self.parts[partNumber];
                if (currentSize != taskDownloadSize) {
                    let message = "download sec size != partSize.  " + currentSize + " " + start + " " + end;
                    let err = self.triggerNetWorkingError(message, partNumber);
                    return reject(err);
                }
                resolve();
            });
            //s3. request的错误监听
            req.on("error", function (err) {
                global.logger.error("[download] s3.request failed,", self.key, partNumber, err);
                delete self.parts[partNumber];
                reject(err);
            });
            req.send();
        });
        return p.then(function () {
            return self.createChunkTask();
        });
    },

    /**
     *   在分段下载上触发一个networking异常，这个异常将会导致重新下载本段
     * @param message    networking异常的描述信息
     * @param partNumber  段号
     * @returns {*|Error}
     */
    triggerNetWorkingError: function (message, partNumber) {
        let self = this;
        let err = new Error(message)
        err = AWS.util.error(err, {code: "NetworkingError", retryable: "true"});
        delete self.completeMap[partNumber];
        self.dataCache.cleanPart(partNumber);
        return err;
    },

    /**
     * 将下载下来的数据写入文件
     */
    writeFile: function () {
        var self = this;
        if (self.isStop) {
            return self.closeFd();
        }

        self.isWriting = true;
        var dataInfo = self.dataCache.shift();
        if (!dataInfo) {
            if (!self.isDownloadOver) {   //带写入的数据已经全部被写入，网络上下载的数据还没有到达.
                self.isWriting = false;
                return;
            } else {
                return self.cleanup();
            }
        }
        fs.write(self.fd, dataInfo.data, 0, dataInfo.length, self.completeMap[dataInfo.partNumber], function (err, bytesWritten, buffer) {
            if (err) {   //写入失败
                global.logger.info("download task write file error...", self.key, err);
                return self.cleanup(err);
            }
            if (bytesWritten !== dataInfo.data.length) {
                return self.cleanup(new Error("the chunk data are not full written."));
            }

            self.completeMap[dataInfo.partNumber] = self.completeMap[dataInfo.partNumber] + dataInfo.data.length;
            self.totalDownloadedBytes += bytesWritten;
            //计算进度
            info = {
                loaded: self.totalDownloadedBytes,
                total: self.size,
                part: dataInfo.partNumber
            };
            self.emit('httpDownloadProgress', [info]);
            self.writeFile();
        });
    },

    abort: function () {
        if (this.size === this.totalDownloadedBytes) {
            logger.info("[download] already download 100% when abort..igore the abort command.", this.key);
            return;
        }
        this.cleanup(AWS.util.error(new Error('Request aborted by user'), {
            code: 'RequestAbortedError',
            retryable: false
        }));
    },


    pause: function () {
        logger.info("[download] pause......... size and totalDownloadedBytes : ", this.key, this.size, this.totalDownloadedBytes)
        if (this.size === this.totalDownloadedBytes) {
            logger.info("[download] already download 100% when pause..igore the pause command.", this.key);
            return;
        }
        this.cleanup(AWS.util.error(new Error("Request paused by user"), {
            code: "RequestPausedError",
            retryable: false
        }));
    },

    /**
     *  移除每段下载req的监听器并对其进行abort。
     *  调用回调函数
     *
     * @api private
     */
    cleanup: function (err) {
        var self = this;
        self.closeFd();
        if (self.isStop)
            return;
        self.isStop = true;
        if ((!err) && self.size != self.totalDownloadedBytes) {
            logger.error("self.size is not equals self.totalDownloadedbytes!:", self.size, self.totalDownloadedBytes);
            err = new Error("size is not equals totalDownloadedbytes");
        }
        self.cleanParts();
        return self.callback(err);
    },

    closeFd: function () {
        var self = this;
        self.isWriting = false;
        if (self.fd) {
            fs.close(self.fd, function (error) {
                if (error) {  //关闭的时候有可能因为文件正在被写入而失败？ 如果是这样在writefile结束后还会再次尝试关闭
                    global.logger.error("[waring] close download file handle error......", self.key);
                } else {
                    global.logger.info("close download file handle OK.....", self.key);
                    self.fd = null;
                }
            })
        }
    },

    cleanParts: function () {
        var self = this;
        AWS.util.each(self.parts, function (partNumber, part) {
            part.removeAllListeners('httpDownloadProgress');
            part.removeAllListeners('complete');
            part.removeAllListeners('httpData');
            part.abort();
        });
        self.parts = {};
    },

    bindServiceObject: function bindServiceObject(params) {
        params = params || {};
        var self = this;

        // bind parameters to new service object
        if (!self.service) {
            self.service = new AWS.S3({
                params: params
            });
        }
        else {
            var config = AWS.util.copy(self.service.config);
            self.service = new self.service.constructor.__super__(config);
            self.service.config.params = AWS.util.merge(self.service.config.params || {}, params);
        }
    },
});

AWS.util.mixin(AWS.S3.ManagedDownload, AWS.SequentialExecutor);
module.exports = AWS.S3.ManagedDownload;