var AWS = require('../core');
var fs = require('fs');
var crypto = require('crypto');

AWS.S3.NodeManagedUpload = AWS.util.inherit({

    constructor : function NodeManagedUpload(options)
    {
        var self = this;
        AWS.SequentialExecutor.call(self);
        self.body = null;
        self.sliceFn = null;
        self.callback = null;
        self.parts = {};
        self.completeInfo = [];
        self.completeMap = {};
        self.fillQueue = function()
        {
            self.callback(new Error('Unsupported body payload ' + typeof self.body));
        };

        self.configure(options);

    },

    /**
     * @api private
     */
    configure : function configure(options)
    {
        options = options || {};
        this.partSize = this.minPartSize;

        if (options.queueSize)
            this.queueSize = options.queueSize;
        if (options.partSize)
            this.partSize = options.partSize;
        if (options.leavePartsOnError)
            this.leavePartsOnError = true;

        if (this.partSize < this.minPartSize) { throw new Error('partSize must be greater than ' + this.minPartSize); }

        this.service = options.service;
        this.bindServiceObject(options.params);
        this.adjustTotalBytes();

        //上传任务号，用于断点续传
        if(options.UploadId){
            this.service.config.params.UploadId = options.UploadId;
            if(options.completeMap){
                this.completeMap=options.completeMap;
            }
            if(options.totalUploaded){
                this.totalUploaded=options.totalUploaded;
                this.totalUploadedBytes=options.totalUploaded;
            }
        }
        else if(options.totalUploaded>0){
            global.logger.error('[uploader]--options missing uploadId---->',options,this.service.config.params.UploadId);
        }
    },

    /**
     * @api private
     */
    leavePartsOnError : false,

    /**
     * @api private
     */
    queueSize : 4,

    /**
     * @api private
     */
    partSize : null,

    /**
     * @readonly
     * @return [Number] the minimum number of bytes for an individual part upload.
     */
    minPartSize : 1024 * 1024 * 5,

    /**
     * @readonly
     * @return [Number] the maximum allowed number of parts in a multipart upload.
     */
    maxTotalParts : 10000,


    send : function(callback)
    {
        var self = this;
        self.callback = callback || function(err)
        {
            if (err)
                throw err;
        };


        if (AWS.util.isNode())
        {

            if (self.totalBytes <= self.partSize){
                self.fillStream();
            }
            else{
                for(var i=0;i<self.queueSize;i++) {
                    self.fillStream();
                }
            }

        }
    },

    /**
     * Aborts a managed upload, including all concurrent upload requests.
     *
     * @note By default, calling this function will cleanup a multipart upload if one was created. To leave the
     *       multipart upload around after aborting a request, configure `leavePartsOnError` to `true` in the
     *       {constructor}.
     * @!macro nobrowser
     * @example Aborting an upload var params = { Bucket: 'bucket', Key: 'key', Body: new Buffer(1024 * 1024 * 25) //
     *          25MB payload }; var upload = s3.upload(params); upload.send(function (err, data) { if (err)
     *          console.log("Error:", err.code, err.message); else console.log(data); });
     *  // abort request in 1 second setTimeout(upload.abort.bind(upload), 1000);
     */
    abort : function()
    {
        this.cleanup(AWS.util.error(new Error('Request aborted by user'), {
            code : 'RequestAbortedError',
            retryable : false
        }));
    },


    /**
     * Pause a managed upload, including all concurrent upload requests.
     */
    pause : function(callback)
    {
        var self = this;
        if (self.failed)
            return;

        AWS.util.each(self.parts, function(partNumber, part)
        {
            part.removeAllListeners('httpUploadProgress');
            part.removeAllListeners('complete');
            part.abort();
        });


        if (self.callback) {
            self.callback(AWS.util.error(new Error("Request paused by user"), {
                code: "RequestPausedError",
                retryable: false
            }));
        }
        self.failed = true;
    }
    ,


    /**
     * @api private
     */
    bindServiceObject : function bindServiceObject(params)
    {
        params = params || {};
        var self = this;

        // bind parameters to new service object
        if (!self.service)
        {
            self.service = new AWS.S3({
                params : params
            });
        }
        else
        {
            var config = AWS.util.copy(self.service.config);
            self.service = new self.service.constructor.__super__(config);
            self.service.config.params = AWS.util.merge(self.service.config.params || {}, params);
        }
    },

    /**
     * @api private
     */
    adjustTotalBytes : function adjustTotalBytes()
    {
        var self = this;

        self.body = self.service.config.params.Body;
        try
        { // try to get totalBytes
            self.totalBytes = fs.lstatSync(self.body).size;
        }
        catch (e)
        {
            console.log(e);
        }

        // try to adjust partSize if we know payload length
        if (self.totalBytes)
        {
            var newPartSize = Math.ceil(self.totalBytes / self.maxTotalParts);
            if (newPartSize > self.partSize)
                self.partSize = newPartSize;
        }
        else
        {
            self.totalBytes = undefined;
        }
    },

    /**
     * @api private
     */
    isDoneChunking : false,

    /**
     * @api private
     */
    partPos : 0,

    /**
     * @api private
     */
    totalChunkedBytes : 0,

    /**
     * @api private
     */
    totalUploadedBytes : 0,

    /**
     * @api private
     * 已上传段总大小
     */
    totalUploaded : 0,

    /**
     * @api private
     */
    totalBytes : undefined,

    /**
     * @api private
     */
    numParts : 0,

    /**
     * @api private
     */
    totalPartNumbers : 0,

    /**
     * @api private
     */
    activeParts : 0,

    /**
     * @api private
     */
    doneParts : 0,

    /**
     * @api private
     */
    parts : null,
    chunks : {},

    /**
     * @api private
     */
    completeInfo : null,

    /**
     * @api private
     */
    completeMap : {},

    /**
     * @api private
     */
    failed : false,

    /**
     * @api private
     */
    multipartReq : null,

    /**
     * @api private
     */
    partBuffer : null,

    /**
     * @api private
     */
    fillStream : function fillStream()
    {
        var self = this;

        if (self.activeParts >= self.queueSize)
            return;


        //算出下载的文件位置
        var start = self.totalChunkedBytes
        var end = (self.totalChunkedBytes + self.partSize < self.totalBytes) ? self.totalChunkedBytes + self.partSize - 1 : self.totalBytes-1
        self.totalChunkedBytes = end + 1;

        //结束
        if(end===self.totalBytes-1){
            self.isDoneChunking=true;

        }

        //只有一片
        if (self.totalBytes <= self.partSize)
        {
            self.isDoneChunking=true;

            if (self.totalBytes > 0)
            {
                self.numParts++;
                self.nextChunk(0,self.totalBytes-1);
            }
        }
        //大于一片
        else
        {
            if(start<=end){
                self.numParts++;
                self.nextChunk(start,end);
            }
        }
    },

    /**
     * @api private
     */
    nextChunk : function nextChunk(start,end)
    {
        var self = this;
        if (self.failed)
            return null;

        var partNumber = ++self.totalPartNumbers;

        if (self.isDoneChunking && partNumber === 1)
        {
            var chunk = fs.createReadStream(self.body,{start: start, end: end}).on('open', function (fd) {
                global.logger.debug('open file:',self.body,{start: start, end: end});
                global.filesNum++;
            }).on('close',function(){
                global.logger.debug('open file:',self.body,{start: start, end: end});
                global.filesNum--;
            }).on('error', function (err) {
                global.logger.error('createReadStream error:',err);
                global.logger.error('global.filesNum:',global.filesNum);
                self.cleanup(err);
            });

            var params = {Body : chunk};
            // if(global.serverEncryption){
            //     params = {Body : chunk,ServerSideEncryption:'aws:kms'};
            // }else{
            //     params = {Body : chunk};
            // }
            var req = self.service.putObject(params);
            self.parts[partNumber] = req;
            req._managedUpload = self;
            req.on('httpUploadProgress', self.progress).send(self.finishSinglePart);
            return null;
        }

        self.activeParts++;
        if (!self.service.config.params.UploadId)
        {

            if (!self.multipartReq)
            { // create multipart
                // var params = {};
                // if(global.serverEncryption){
                //     self.multipartReq = self.service.createMultipartUpload({ServerSideEncryption:'aws:kms'});
                // }else{
                //     self.multipartReq = self.service.createMultipartUpload();
                // }
                self.multipartReq = self.service.createMultipartUpload();

                self.multipartReq.on('success', function(resp)
                {
                    var uploadId=resp.data.UploadId;
                    if(!self.service.config.params.UploadId){
                        self.service.config.params.UploadId = uploadId;
                        self.emit('successCreateMultipartUpload', [uploadId]);
                        self.multipartReq = null;
                    }
                });
                self.queueChunks(start,end, partNumber);
                self.multipartReq.on('error', function(err)
                {
                    global.logger.error('[Error]-->multipartReq--->>>',err,self.completeMap,self.service.config.params.UploadId);
                    self.cleanup(err);
                });
                self.multipartReq.send();
            }
            else
            {
                self.queueChunks(start,end, partNumber);
            }
        }
        else
        { // multipart is created, just send
            self.uploadPart(start,end, partNumber);
        }
    },

    /**
     * @api private
     */
    uploadPart : function uploadPart(start,end, partNumber)
    {
        var self = this;

        var partInfo = {
            ETag : null,
            PartNumber : partNumber
        };
        self.completeInfo.push(partInfo);

        //该段已经上传
        if(self.completeMap[partNumber]){

            setTimeout(function(){

                partInfo.ETag = self.completeMap[partNumber];
                self.activeParts--;
                self.doneParts++;

                if (self.isDoneChunking && self.doneParts === self.numParts)
                {
                    self.finishMultiPart();
                }
                else
                {
                    self.fillStream.call(self);
                }
            },100);
        }
        //没有上传
        else{

            var chunk = fs.createReadStream(self.body,{start: start, end: end}).on('open', function (fd) {
                global.logger.debug('open file:',self.body,{start: start, end: end});
                global.filesNum++;
            }).on('close',function(){
                global.logger.debug('close file:',self.body,{start: start, end: end});
                global.filesNum--;
            }).on('error', function (err) {
                global.logger.error('createReadStream error:',err,self.service.config.params.UploadId);
                global.logger.error('global.filesNum:',global.filesNum);
                return self.cleanup(err);
            });
            if(global.isMD5check === 'Y'){
                var md5num = crypto.createHash('md5');
                var chunk1 = fs.createReadStream(self.body,{start: start, end: end}).on('data', function (data) {
                    md5num.update(data);
                }).on('end',function(){
                    var partMD5 = md5num.digest('base64');
                    var contentLength = self.partSize;
                    if(partNumber > self.totalBytes/self.partSize){
                        contentLength = self.totalBytes-(self.partSize*(partNumber-1));
                    }
                    var partParams = {
                        Body : chunk,
                        ContentLength : contentLength,
                        PartNumber : partNumber,
                        ContentMD5:partMD5
                    };
                    var req = self.service.uploadPart(partParams);
                    self.parts[partNumber] = req;
                    self.chunks[partNumber] = chunk;
                    req._lastUploadedBytes = 0;
                    req._managedUpload = self;
                    req.on('httpUploadProgress', self.progress);
                    req.send(function(err, data)
                    {
                        chunk.destroy();
                        chunk1.destroy();
                        delete self.parts[partParams.PartNumber];
                        delete self.chunks[partParams.PartNumber];
                        self.activeParts--;

                        if (!err && (!data || !data.ETag))
                        {
                            var message = 'No access to ETag property on response.';
                            if (AWS.util.isBrowser())
                            {
                                message += ' Check CORS configuration to expose ETag header.';
                            }

                            err = AWS.util.error(new Error(message), {
                                code : 'ETagMissing',
                                retryable : false
                            });
                        }
                        if (err)
                            return self.cleanup(err);

                        partInfo.ETag = data.ETag;
                        self.completeMap[partInfo.PartNumber] = partInfo.ETag;
                        self.totalUploaded+=contentLength;

                        self.emit('partComplete',[{partNumber:partInfo.PartNumber,etag:partInfo.ETag,uploaded:self.totalUploaded,uploadId:self.service.config.params.UploadId}]);
                        self.doneParts++;

                        if (self.isDoneChunking && self.doneParts === self.numParts)
                        {
                            self.finishMultiPart();
                        }
                        else
                        {
                            self.fillStream.call(self);
                        }
                    });
                })
            }else{
                var contentLength = self.partSize;
                if(partNumber > self.totalBytes/self.partSize){
                    contentLength = self.totalBytes-(self.partSize*(partNumber-1));
                }
                var partParams = {
                    Body : chunk,
                    ContentLength : contentLength,
                    PartNumber : partNumber
                };
                var req = self.service.uploadPart(partParams);
                self.parts[partNumber] = req;
                self.chunks[partNumber] = chunk;
                req._lastUploadedBytes = 0;
                req._managedUpload = self;
                req.on('httpUploadProgress', self.progress);
                req.send(function(err, data)
                {
                    chunk.destroy();
                    delete self.parts[partParams.PartNumber];
                    delete self.chunks[partParams.PartNumber];
                    self.activeParts--;

                    if (!err && (!data || !data.ETag))
                    {
                        var message = 'No access to ETag property on response.';
                        if (AWS.util.isBrowser())
                        {
                            message += ' Check CORS configuration to expose ETag header.';
                        }

                        err = AWS.util.error(new Error(message), {
                            code : 'ETagMissing',
                            retryable : false
                        });
                    }
                    if (err)
                        return self.cleanup(err);

                    partInfo.ETag = data.ETag;
                    self.completeMap[partInfo.PartNumber]=partInfo.ETag;
                    self.totalUploaded+=contentLength;

                    self.emit('partComplete',[{partNumber:partInfo.PartNumber,etag:partInfo.ETag,uploaded:self.totalUploaded,uploadId:self.service.config.params.UploadId}]);
                    self.doneParts++;

                    if (self.isDoneChunking && self.doneParts === self.numParts)
                    {
                        self.finishMultiPart();
                    }
                    else
                    {
                        self.fillStream.call(self);
                    }
                });
            }
        }

    },

    /**
     * @api private
     */
    queueChunks : function queueChunks(start,end, partNumber)
    {
        var self = this;
        self.multipartReq.on('success', function()
        {
            self.uploadPart(start,end, partNumber);
        });
    },

    /**
     * @api private
     */
    cleanup : function cleanup(err)
    {
        var self = this;
        if (self.failed)
            return;

        if (self.service.config.params.UploadId && !self.leavePartsOnError)
        {
            self.service.abortMultipartUpload().send();
        }

        AWS.util.each(self.chunks, function(partNumber, chunk)
        {
            chunk.destroy();
        });
        AWS.util.each(self.parts, function(partNumber, part)
        {
            part.removeAllListeners('complete');
            part.abort();
        });

        self.parts = {};
        self.callback(err);
        self.failed = true;
    },


    /**
     * @api private
     */
    listParts : function listParts()
    {
        var self = this;

        self.service.listParts(function(err, data) {
            if (err)
                console.log(err)
            else{
                if(data.Parts){
                    for(var i=0;i<data.Parts.length;i++){
                        var partInfo=data.Parts[i];
                        self.completeMap[partInfo.PartNumber]=partInfo.ETag;
                        self.totalUploadedBytes+=partInfo.Size;
                    }
                }
            }
        });
    },


    /**
     * @api private
     */
    finishMultiPart : function finishMultiPart()
    {
        var self = this;
        var completeParams = {
            MultipartUpload : {
                Parts : self.completeInfo
            }
        };
        self.service.completeMultipartUpload(completeParams, function(err, data)
        {
            if (err){
                global.logger.error('completeMultipartUpload error:',err,self.service.config.params.UploadId);
                return self.cleanup(err);
            } else
                self.callback(err, data);

        });
    },

    /**
     * @api private
     */
    finishSinglePart : function finishSinglePart(err, data)
    {
        var upload = this.request._managedUpload;
        var httpReq = this.request.httpRequest;
        var url = AWS.util.urlFormat(httpReq.endpoint);
        if (err)
            return upload.callback(err);
        data.Location = url.substr(0, url.length - 1) + httpReq.path;
        upload.callback(err, data);
    },

    /**
     * @api private
     */
    progress : function progress(info)
    {
        var upload = this._managedUpload;
        if (this.operation === 'putObject')
        {
            info.part = 1;
        }
        else
        {
            upload.totalUploadedBytes += info.loaded - this._lastUploadedBytes;
            this._lastUploadedBytes = info.loaded;
            info = {
                loaded : upload.totalUploadedBytes,
                total : upload.totalBytes,
                part : this.params.PartNumber
            };
        }
        upload.emit('httpUploadProgress', [ info ]);
    }
});

AWS.util.mixin(AWS.S3.NodeManagedUpload, AWS.SequentialExecutor);
module.exports = AWS.S3.NodeManagedUpload;
